#------------------------------------------------------------
"""
Send Morse

New version May 12, 2015

Build the complete utterance directly in a temp WAV file
using writeframes, then play the wav file.
Each dit, dah and space gets built as a small array, then
the array joins the wav file.
When done, play.

June 2017: just for testing, see if Morse can be emotional.
Treat numbers as pitch indicators instead of numbers.


"""
#------------------------------------------------------------

from Tkinter import *
import tkFileDialog

import os
import string
import array
import wave
import math
import random
import winsound


#------------------------------------------------------------

# Global vars
global TextFileName
global WPM
global CodeDict
global DitArray,DahArray,IPIArray,ICIArray,IWIArray
global IPIMade,ICIMade,IWIMade
global Frequency

TextFileName = ''
WaveFileName='wavetemp.wav'
Rate = 44100.0
Frequency = 1000.0
TwoPi=2.0*math.pi
UpRamp=30.0
DnRamp=30.0
Level = 32676.0 / 10.0


# Figure elements from WPM:
WPM=13
DitMsec = 1000.0/(WPM * 50.0 / 60.0)
DahMsec = DitMsec * 3.0
IPIMsec = DitMsec        # inter-pulse interval, same as dit
ICIMsec = DahMsec        # inter-character interval, same as dah
IWIMsec = DitMsec * 7.0  # inter-word interval, 7 * dit

IPIMade,ICIMade,IWIMade = 0,0,0



#------------------------------------------------------------

# Freqs by number for intonation test
FreqsByNumber = {

"0" : 500,
"1" : 600,
"2" : 700,
"3" : 800,
"4" : 900,
"5" : 1050,
"6" : 1175,
"7" : 1300,
"8" : 1450,
"9" : 1600
}



#------------------------------------------------------------

# Set up the code array as a dictionary.

CodeDict = {
 " ":" ",
 "A":".-",
 "B":"-...",
 "C":"-.-.",
 "D":"-..",
 "E":".",
 "F":".-..",
 "G":"--.",
 "H":"....",
 "I":"..",
 "J":".---",
 "K":"-.-",
 "L":".-..",
 "M":"--",
 "N":"-.",
 "O":"---",
 "P":".--.",
 "Q":"--.-",
 "R":".-.",
 "S":"...",
 "T":"-",
 "U":"..-",
 "V":"...-",
 "W":".--",
 "X":"-..-",
 "Y":"-.--",
 "Z":"--..",
 ".":".-.-.-",
 ",":"--..--",
 "?":"..--..",
 "0":"-----",
 "1":".----",
 "2":"..---",
 "3":"...--",
 "4":"....-",
 "5":".....",
 "6":"-....",
 "7":"--...",
 "8":"---..",
 "9":"----."
}

#
'''
 Separately, a list of letters repeated in proportion to
 frequency in language.
 These are fudged to put K and W higher than normal.

 Original freqs (which are NOT proper English freqs)
LettersByFreq=
 "E": 12,
 "T": 9,
 "A": 8.
 "O": 7,
 "I": 7,
 "N": 6,
 "S": 6,
 "R": 6,
 "H": 5,
 "W": 4,
 "K": 4,
 "D": 4,
 "L": 3,
 "U": 3,
 "C": 2.
 "M": 2,
 "F": 2,
 "Y": 2,
 "G": 2,
 "P": 1,
 "B": 1,
 "V": 1,
 "X": 1,
 "Q": 1,
 "J": 1,
 "Z": 1,
 ".": 1.
 ",": 1,
 "?": 1,
 "0": 1,
 "1": 1,
 "2": 1,
 "3": 1,
 "4": 1,
 "5": 1,
 "6": 1,
 "7": 1,
 "8": 1,
 "9": 1
}
 End source table

'''
LettersByFreq=[
 'E','E','E','E','E','E','E','E','E','E','E','E',
 'T','T','T','T','T','T','T','T','T',
 'A','A','A','A','A','A','A','A',
 'O','O','O','O','O','O','O',
 'I','I','I','I','I','I','I',
 'N','N','N','N','N','N',
 'S','S','S','S','S','S',
 'R','R','R','R','R','R',
 'H','H','H','H','H',
 'W','W','W','W',
 'K','K','K','K',
 'D','D','D','D',
 'L','L','L',
 'U','U','U',
 'C','C',
 'M','M',
 'F','F',
 'Y','Y',
 'G','G',
 'P',
 'B',
 'V',
 'X',
 'Q',
 'J',
 'Z',
 '.',
 ',',
 '?',
 '0',
 '1',
 '2',
 '3',
 '4',
 '5',
 '6',
 '7',
 '8',
 '9',
]


#------------------------------------------------------------

def BuildDit():

    global DitArray,Frequency

    # cycles / buffer = (cycles / second) * (seconds / buffer).
    CyclesPerBuffer = Frequency * (DitMsec / 1000.0)

    # Samples / Cycle = (Samples / sec) / (Cycles / sec)
    SamplesPerCycle = Rate / Frequency

    # Now get the whole number of points in this number of cycles.
    # (Samples / Buffer) =  (Samples / Cycle) * (Cycles / Buffer)
    BufLen = int(SamplesPerCycle * CyclesPerBuffer)
    # should be 3540

    DitArray = array.array('h',[0]*BufLen) # Array of zeroes

    # Figure the angle increment for each sampling point:
    #    Radians / Sample = radians/Cycle * Cycles / Sample.
    # or Radians / Sample = TwoPi         / (SamplesPerCycle)
    RadStep = TwoPi / float(SamplesPerCycle)
    # Figure where in the buffer to start and stop, from the msec.
    # Points = Msec * (Points/Msec)
    # Note! These are points, not bytes.
    Ramp=0.0
    FrontRamp=0.0
    BackRamp=0.0

    for i in range(BufLen):
        # Get radians corresponding to this index:
        Rads = float(i) * RadStep

        # Set ramp: arbitrarily ramp up for first 30 points, down last 30 pts.
        Ramp = 1.0
        FrontRamp = float(i)
        BackRamp = float(BufLen - i)
        if   (FrontRamp < UpRamp): Ramp = FrontRamp / UpRamp
        elif (BackRamp < DnRamp):  Ramp = BackRamp / DnRamp

        # Put sin value in the buffer (at int index)
        DitArray[i] =  int(Level * Ramp * math.sin(Rads))

    # Write this to the Wavefile
    fpWave.writeframes(DitArray.tostring())

#------------------------------------------------------------

def BuildDah():
    global DahArray,Frequency

    # cycles / buffer = (cycles / second) * (seconds / buffer).
    CyclesPerBuffer = Frequency * (DahMsec / 1000.0)

    # Samples / Cycle = (Samples / sec) / (Cycles / sec)
    SamplesPerCycle = Rate / Frequency

    # Now get the whole number of points in this number of cycles.
    # (Samples / Buffer) =  (Samples / Cycle) * (Cycles / Buffer)
    BufLen = int(SamplesPerCycle * CyclesPerBuffer)

    DahArray = array.array('h',[0]*BufLen) # Array of zeroes

    # Figure the angle increment for each sampling point:
    #    Radians / Sample = radians/Cycle * Cycles / Sample.
    # or Radians / Sample = TwoPi         / (SamplesPerCycle)
    RadStep = TwoPi / float(SamplesPerCycle)
    # Figure where in the buffer to start and stop, from the msec.
    # Points = Msec * (Points/Msec)
    # Note! These are points, not bytes.
    Ramp=0.0
    FrontRamp=0.0
    BackRamp=0.0

    for i in range(BufLen):
        # Get radians corresponding to this index:
        Rads = float(i) * RadStep

        # Set ramp: arbitrarily ramp up for first 30 points, down last 30 pts.
        Ramp = 1.0
        FrontRamp = float(i)
        BackRamp = float(BufLen - i)
        if   (FrontRamp < UpRamp): Ramp = FrontRamp / UpRamp
        elif (BackRamp < DnRamp):  Ramp = BackRamp / DnRamp

        # Put sin value in the buffer (at int index)
        DahArray[i] = int(Level * Ramp * math.sin(Rads))
    # Write this to the Wavefile
    fpWave.writeframes(DahArray.tostring())

#------------------------------------------------------------

def BuildIPI():
    global IPIArray,IPIMade
    if not IPIMade:
        # No need for cycles and ramps here, just make zeros for the required length
        BufLen = int(Rate * IPIMsec / 1000.0)
        IPIArray = array.array('h',[0]*BufLen) # Array of zeroes
        # And that's all we need here
        IPIMade=1
    # Write this to the Wavefile
    fpWave.writeframes(IPIArray.tostring())

#------------------------------------------------------------

def BuildICI():
    global ICIArray,ICIMade
    if not ICIMade:
        # No need for cycles and ramps here, just make zeros for the required length
        BufLen = int(Rate * ICIMsec / 1000.0)
        ICIArray = array.array('h',[0]*BufLen) # Array of zeroes
        # And that's all we need here
        ICIMade=1
    # Write this to the Wavefile
    fpWave.writeframes(ICIArray.tostring())

#------------------------------------------------------------

def BuildIWI():
    global IWIArray,IWIMade
    if not IWIMade:
        # No need for cycles and ramps here, just make zeros for the required length
        BufLen = int(Rate * IWIMsec / 1000.0)
        IWIArray = array.array('h',[0]*BufLen) # Array of zeroes
        # And that's all we need here
        IWIMade=1
    # Write this to the Wavefile
    fpWave.writeframes(IWIArray.tostring())

#------------------------------------------------------------

def ProcessOneLetter(LetterArg):

    global Frequency

    # Always put one interletter space before each letter.
    BuildICI()

    # For intonation try:
    if LetterArg.isdigit():
        Frequency = FreqsByNumber[LetterArg]
    else:
        # Look up the letter in the code dict.
        try:
            Pattern = CodeDict[LetterArg]
        except: # anything unknown, make a space.
            Pattern = " "

        # Take the Pattern and turn it into moves of the Key and Sounder.
        for OneDotOrDash in Pattern:
            if OneDotOrDash == ".":
                BuildDit()
                BuildIPI()
            elif OneDotOrDash == "-":
                BuildDah()
                BuildIPI()
            elif OneDotOrDash == " ": # Space between words. Because we've already
                # moved by 3 dits before the text-letter, use 4 instead of standard 7.
                BuildIWI()
    # end actions when not just setting freq

#------------------------------------------------------------

def ReadFile(FileArg):
    ReadString = ""
    tf = 0
    try:
        tf = open(FileArg,"r")
    except:
        raise 'Text file not found!'
    try:
        TryString = tf.read()
    except:
        raise "Text file not readable!"
    # Make the real string all upper. (Critical!)
    ReadString = string.upper(TryString)
    tf.close()
    return ReadString

#------------------------------------------------------------

def RunMorse():
    global TextFileName
    global WPM
    global fpWave

    FullString = ReadFile(TextFileName)

    StringIndex = 0

    # Make the temp wave file
    fpWave=wave.open(WaveFileName,'wb')
    if not fpWave:
        raise Exception("Wave file failed to open.")
    fpWave.setnchannels(1)
    fpWave.setsampwidth(2)
    fpWave.setframerate(Rate)

    for i in range(len(FullString)):
        OneChar = FullString[i]
        ProcessOneLetter(OneChar)

    fpWave.close()
    # Now that the wavefile is built, play it.

    #app.ShowStatus("Sending....")
    #winsound.PlaySound(WaveFileName,winsound.SND_FILENAME)


#------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):

        self.master = master
        master.title('Morse sender')

        self.StatusEntry = Entry(self.master,width=60)
        self.StatusEntry.grid(row=0,column=0,columnspan=6,pady=6)
        self.StatusEntry.insert(0,"Set WPM if needed, then get TXT.")


        Label(master, text="Words per minute:" ).grid(row=1, column=0)
        self.WPMEntry = Entry(self.master,width=6)
        self.WPMEntry.grid(row=1,column=1,pady=6)
        self.WPMEntry.insert(0,"13")

        self.buttonRandRaw = Button(master, text="Random 1", command=self.HandleGetRandRaw)
        self.buttonRandRaw.grid(row=1, column=2,pady=6)

        self.buttonRandFreq = Button(master, text="Random 2", command=self.HandleGetRandFreq)
        self.buttonRandFreq.grid(row=1, column=3,pady=6)

        self.buttonGet = Button(master, text="Get TXT", command=self.HandleGetTXT)
        self.buttonGet.grid(row=1, column=4,pady=6)

        # buttons for Go, cancel.
        self.buttonGo = Button(master, text="Send", background="red", command=self.HandleSend)
        self.buttonGo.grid(row=1, column=5,pady=6)

    # - - - - - - - - - - - - - - - - - -


    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def HandleSend(self):
        global TextFileName
        global WPM
        global DitMsec, DahMsec, IPIMsec, ICIMsec, IWIMsec

        # Fetch the wpm
        try:
          WPM =  float(self.WPMEntry.get())
        except:
          WPM = 13.0
        DitMsec = 1000.0/(WPM * 50.0 / 60.0)
        DahMsec = DitMsec * 3.0
        IPIMsec = DitMsec        # inter-pulse interval, same as dit
        ICIMsec = DahMsec        # inter-character interval, same as dah
        IWIMsec = DitMsec * 7.0  # inter-word interval, 7 * dit

        self.ShowStatus("Processing text....")
        RunMorse()
        self.ShowStatus("Done. Hit Send again or get new TXT.")

    # - - - - - - - - - - - - - - - - - -

    def HandleGetTXT(self):
        # Pop up file box to get name.
        global TextFileName
        TextFileName=tkFileDialog.askopenfilename(filetypes=[("TXT", "txt")] )

        if not TextFileName:
            return 0
        self.ShowStatus("Now hit Send.")


    # - - - - - - - - - - - - - - - - - -

    def HandleGetRandRaw(self):
        # Create a random text about 5 minutes long.
        global TextFileName
        self.ShowStatus("Building raw random text....")
        # Write a rand txt file, provide its name just as if you'd selected
        RandString=''
        TotalWords = int( random.uniform(0.8,1.2)*5.0*WPM )
        for i in range(TotalWords):
            wordlen = int( random.uniform(-2,2) )+5
            for j in range(wordlen):
                letterindex = random.randrange(len(CodeDict))
                letter=CodeDict.keys()[letterindex]
                if letter in ('.' , ',' , '?'): # Punctuation
                    if j==0:
                        # if punct would be at start of word, skip entirely
                        break
                    else:
                        # if punct not at start, let it enter the string
                        #and break for space after the punct
                        RandString = RandString+letter
                        break
                else:
                    # Not punct, just run normally
                    RandString = RandString+letter

                # If we picked punctuation, cut the 'word' here and let a space follow


            RandString = RandString + ' ' # space after each word

        # Keep a sequence of Rand%nn files up to 99.
        for N in range(0,100): # 100 should be enough; also avoids infinite loop.
            TryName = "RandRaw-%.2d.TXT" % N
            if os.path.isfile(TryName): continue # Exists. Keep trying until 99

            else: #  Doesn't exist. Make one and stop
                fpRand=open(TryName,'wt')
                fpRand.write(RandString)
                fpRand.close()
                TextFileName=TryName # set global name so this will act like GetTXT
                break
        # Name-try loop, ended by making one


        self.ShowStatus("Now hit Send.")

    # - - - - - - - - - - - - - - - - - -

    def HandleGetRandFreq(self):
        # Create a random text about 5 minutes long.
        global TextFileName
        self.ShowStatus("Building language-like random text....")
        # Write a rand txt file, provide its name just as if you'd selected
        RandString=''

        TotalWords = int( random.uniform(0.8,1.2)*5.0*WPM )

        # Use freq table to prefer the most common.
        for i in range(TotalWords):
            wordlen = int( random.uniform(-2,2) )+5
            for j in range(wordlen):
                # Use the freq-shaped list to get a letter
                # Remember that we're just getting letters here, not dits and dahs
                letterindex = random.randrange(len(LettersByFreq))
                letter = LettersByFreq[letterindex]
                if letter in ('.' , ',' , '?'): # Punctuation
                    if j==0:
                        # if punct would be at start of word, skip entirely
                        break
                    else:
                        # if punct not at start, let it enter the string
                        #and break for space after the punct
                        RandString = RandString+letter
                        break
                else:
                    # Not punct, just run normally
                    RandString = RandString+letter

                # If we picked punctuation, cut the 'word' here and let a space follow


            RandString = RandString + ' ' # space after each word

        # Keep a sequence of Rand%nn files up to 99.
        for N in range(0,100): # 100 should be enough; also avoids infinite loop.
            TryName = "RandFreq-%.2d.TXT" % N
            if os.path.isfile(TryName): continue # Exists. Keep trying until 99

            else: #  Doesn't exist. Make one and stop
                fpRand=open(TryName,'wt')
                fpRand.write(RandString)
                fpRand.close()
                TextFileName=TryName # set global name so this will act like GetTXT
                break
        # Name-try loop, ended by making one


        self.ShowStatus("Now hit Send.")

    # - - - - - - - - - - - - - - - - - -

    def quit(self):
        self.master.destroy()

#------------------------------------------------------------
# Activate the Win loop, which will be broken by any button.
#------------------------------------------------------------

root = Tk()
app = App(root, '')
root.mainloop()

#------------------------------------------------------------
